﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Common;
using System.Transactions;
using System.Data.SqlClient;

namespace HeuristicLab.Services.Deployment.DataAccess {
  public class PluginStore {

    private PluginStoreClassesDataContext ctx;
    public PluginStore() {
      ctx = new PluginStoreClassesDataContext();
    }

    public IEnumerable<ProductDescription> Products {
      get {
        return from p in ctx.Products
               let plugins = from pair in ctx.ProductPlugins
                             from plugin in ctx.Plugins
                             where pair.ProductId == p.Id
                             where plugin.Id == pair.PluginId
                             select plugin
               select MakeProductDescription(p, plugins);
      }
    }

    private ProductDescription MakeProductDescription(Product p, IQueryable<Plugin> plugins) {
      var desc = new ProductDescription(p.Id, p.Name, new Version(p.Version), from plugin in plugins
                                                                              select MakePluginDescription(plugin));
      return desc;
    }

    private PluginDescription MakePluginDescription(Plugin plugin) {
      var desc = new PluginDescription(plugin.Id, plugin.Name, new Version(plugin.Version), from dep in GetDependencies(plugin)
                                                                                            select MakePluginDescription(dep));
      return desc;
    }

    private IEnumerable<Plugin> GetDependencies(Plugin plugin) {
      return from pair in ctx.Dependencies
             from dependency in ctx.Plugins
             where pair.PluginId == plugin.Id
             where pair.DependencyId == dependency.Id
             select dependency;
    }

    public IEnumerable<PluginDescription> Plugins {
      get {
        return from plugin in ctx.Plugins
               select MakePluginDescription(plugin);
      }
    }

    public byte[] PluginFile(PluginDescription pluginDescription) {
      return (from file in ctx.PluginPackages
              where file.PluginId == pluginDescription.Id
              select file.Data.ToArray()).Single();
    }

    public void Persist(PluginDescription pluginDescription, byte[] pluginPackage) {
      try {
        using (var transaction = new TransactionScope()) {
          InsertOrUpdatePlugin(pluginDescription);
          InsertOrUpdatePluginPackage(pluginDescription, pluginPackage);
          ctx.SubmitChanges();
          transaction.Complete();
        }
      }
      catch (SqlException ex) {
        throw new ArgumentException("Something went wrong while trying to persist plugin", ex);
      }
    }

    public void Persist(ProductDescription product) {
      using (var transaction = new TransactionScope()) {
        InsertOrUpdateProduct(product);
        foreach (var plugin in product.Plugins) {
          InsertOrUpdatePlugin(plugin);
        }
        ctx.SubmitChanges();
        transaction.Complete();
      }
    }

    private void InsertOrUpdatePluginPackage(PluginDescription pluginDescription, byte[] pluginPackage) {
      var existing = from p in ctx.PluginPackages
                     where p.PluginId == pluginDescription.Id
                     select p;

      if (existing.Count() == 0) {
        ctx.PluginPackages.InsertOnSubmit(MakePluginPackage(pluginDescription, pluginPackage));
      } else {
        var exisingPackage = existing.Single();
        exisingPackage.Data = pluginPackage;
      }
    }

    private void InsertOrUpdateProduct(ProductDescription product) {
      var existing = from p in ctx.Products
                     where p.Id == product.Id
                     select p;
      if (existing.Count() == 0) {
        var newEntity = MakeProductFromDescription(product);
        ctx.Products.InsertOnSubmit(newEntity);
        // submit and write back db generated key
        ctx.SubmitChanges();
        product.Id = newEntity.Id;
      } else {
        var existingProduct = existing.Single();
        existingProduct.Name = product.Name;
        existingProduct.Version = product.Version.ToString();
      }
      foreach (var plugin in product.Plugins) {
        InsertOrUpdateProductPlugin(plugin, product);
        InsertOrUpdatePlugin(plugin);
      }
    }

    private void InsertOrUpdateProductPlugin(PluginDescription plugin, ProductDescription product) {
      var existing = from pair in ctx.ProductPlugins
                     where pair.PluginId == plugin.Id
                     where pair.ProductId == product.Id
                     select pair;
      if (existing.Count() != 0) {
        var newEntity = new ProductPlugin();
        newEntity.PluginId = plugin.Id;
        newEntity.ProductId = product.Id;
        ctx.ProductPlugins.InsertOnSubmit(newEntity);
      }
    }

    private void InsertOrUpdatePlugin(PluginDescription plugin) {
      var existing = from p in ctx.Plugins
                     where p.Name == plugin.Name
                     where p.Version == plugin.Version.ToString()
                     select p;
      if (existing.Count() == 0) {
        var newEntity = MakePluginFromDescription(plugin);
        ctx.Plugins.InsertOnSubmit(newEntity);
        // submit and write back db generated key
        ctx.SubmitChanges();
        plugin.Id = newEntity.Id;
      } else {
        var existingPlugin = existing.Single();
        plugin.Id = existingPlugin.Id;
      }
      foreach (var dependency in plugin.Dependencies) {
        InsertOrUpdateDependency(plugin, dependency);
        InsertOrUpdatePlugin(dependency);
      }
    }

    private void InsertOrUpdateDependency(PluginDescription plugin, PluginDescription dependency) {
      var existing = from pair in ctx.Dependencies
                     where pair.DependencyId == dependency.Id
                     where pair.PluginId == plugin.Id
                     select pair;
      if (existing.Count() == 0) {
        var newEntity = new Dependency();
        newEntity.PluginId = plugin.Id;
        newEntity.DependencyId = dependency.Id;
        ctx.Dependencies.InsertOnSubmit(newEntity);
      }
    }

    private Plugin MakePluginFromDescription(PluginDescription pluginDescription) {
      var plugin = new Plugin();
      plugin.Id = pluginDescription.Id;
      plugin.Name = pluginDescription.Name;
      plugin.Version = pluginDescription.Version.ToString();
      return plugin;
    }

    private Product MakeProductFromDescription(ProductDescription desc) {
      var product = new Product();
      product.Id = desc.Id;
      product.Name = desc.Name;
      product.Version = desc.Version.ToString();
      return product;
    }

    private PluginPackage MakePluginPackage(PluginDescription pluginDescription, byte[] pluginPackage) {
      var package = new PluginPackage();
      package.Data = pluginPackage;
      package.PluginId = pluginDescription.Id;
      package.FileName = string.Empty;
      return package;
    }
  }
}
